@@ -1,14 +1,14 @@
 #[macro_use]
 extern crate std;
 
-use bitarray::BitArray;
+use bitarray::{BitArray, Hamming};
 use hgg::Hgg;
 use rand::{Rng, SeedableRng};
-use space::Knn;
+use space::{Knn, KnnInsert};
 
 #[test]
 fn random_insertion_stats() {
-    let mut hgg: Hgg<BitArray<32>, ()> = Hgg::new().insert_knn(100);
+    let mut hgg: Hgg<Hamming, BitArray<32>, ()> = Hgg::default().insert_knn(100);
 
     // Use a PRNG with good statistical properties for generating 64-bit numbers.
     let mut rng = rand_xoshiro::Xoshiro256PlusPlus::seed_from_u64(0);
